/*

*/
package pipeline;
import scene.Vertex;
import scene.primitives.Triangle;
import framebuffer.*;

import java.awt.Color;

/**
   Rasterize each triangle into pixels in the FrameBuffer's viewport.

   This rasterization algorithm is based on
     "Fundamentals of Computer Graphics", 3rd Edition,
      by Peter Shirley, pages 163-165.
*/
public class Rasterize_T
{
   public static boolean debug = false;

   /**
      Rasterize each triangle into pixels in the framebuffer.
   */
   public static void rasterize(Triangle t, FrameBuffer fb)
   {
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      Vertex v0 = t.vertexList.get(0);
      Vertex v1 = t.vertexList.get(1);
      Vertex v2 = t.vertexList.get(2);

      // Get the colors from the three vertices.
      double r0 = v0.r;
      double g0 = v0.g;
      double b0 = v0.b;
      double r1 = v1.r;
      double g1 = v1.g;
      double b1 = v1.b;
      double r2 = v2.r;
      double g2 = v2.g;
      double b2 = v2.b;

      // Get the coordinates from the three vertices.
      double x0 = v0.x;
      double y0 = v0.y;
      double z0 = v0.z;
      double x1 = v1.x;
      double y1 = v1.y;
      double z1 = v1.z;
      double x2 = v2.x;
      double y2 = v2.y;
      double z2 = v2.z;

      // Cull degenerate triangles (triangles that projected onto a line).
      // NOTE: This culling must be done after the projection stage.
      //       In view coordinates we determine front and back
      //       facing triangles, but we don't find the ones that
      //       project to lines until we have done the projection.
      double area = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);
      if ( 0.0001 > Math.abs(area) )
         return;

      // Recall that this triangle's viewport coordinates are in the intervals
      //     x in (-0.5, vpWidth  - 0.5)
      //     y in (-0.5, vpHeight - 0.5)
      // We want to "sample" the triangle (in viewport coordinates)
      // at the pixel locations with integer coordinates between
      //     0 and vpWidth - 1
      //     0 and vpHeight - 1
      // Our first step is to find a "pixel bounding box" for the triangle.
      // Each vertex of the triangle is rounded to integer coordinates, and
      // then we find the left-most, right-most, highest, and lowest integer
      // coordinates that bound the triangle's vertices.

      // Find the greatest integer less than or equal to x0, x1, and x2.
      int xMin = (int)x0;
      if (x1 < xMin) xMin = (int)x1;
      if (x2 < xMin) xMin = (int)x2;
      if (xMin == -1) xMin++;

      // Find the greatest integer less than or equal to y0, y1, and y2.
      int yMin = (int)y0;
      if (y1 < yMin) yMin = (int)y1;
      if (y2 < yMin) yMin = (int)y2;
      if (yMin == -1) yMin++;

      // Find the least integer greater than or equal to x0, x1, and x2.
      int xMax = (int)Math.ceil(x0);
      if (x1 > xMax) xMax = (int)Math.ceil(x1);
      if (x2 > xMax) xMax = (int)Math.ceil(x2);
      if (xMax == fb.getWidthVP()) xMax--;

      // Find the least integer greater than or equal to y0, y1, and y2.
      int yMax = (int)Math.ceil(y0);
      if (y1 > yMax) yMax = (int)Math.ceil(y1);
      if (y2 > yMax) yMax = (int)Math.ceil(y2);
      if (yMax == fb.getHeightVP()) yMax--;
      //System.err.printf("xMin=% 4d, xMax=% 4d, yMin=% 4d, yMax=% 4d\n", xMin,xMax,yMin,yMax);

    //for (int y = yMin;   y <  yMax; y++) // bottom to top
      for (int y = yMax-1; y >= yMin; y--) // top to bottom
      {
         for (int x = xMin; x < xMax; x++)  /// left to right
         {
            //  f01(x,y) = (y0-y1)*(x) + (x1-x0)*(y) + x0*y1 - x1*y0
            //  f12(x,y) = (y1-y2)*(x) + (x2-x1)*(y) + x1*y2 - x2*y1
            //  f20(x,y) = (y2-y0)*(x) + (x0-x2)*(y) + x2*y0 - x0*y2

            //  alpha = f12(x,y) / f12(x0,y0)
            //  beta  = f20(x,y) / f20(x1,y1)
            //  gamma = f01(x,y) / f01(x2,y2)

            double alpha = ( (y1-y2)*(x)  + (x2-x1)*(y)  + x1*y2 - x2*y1 )
                         / ( (y1-y2)*(x0) + (x2-x1)*(y0) + x1*y2 - x2*y1 );
            double beta  = ( (y2-y0)*(x)  + (x0-x2)*(y)  + x2*y0 - x0*y2 )
                         / ( (y2-y0)*(x1) + (x0-x2)*(y1) + x2*y0 - x0*y2 );
            double gamma = ( (y0-y1)*(x)  + (x1-x0)*(y)  + x0*y1 - x1*y0 )
                         / ( (y0-y1)*(x2) + (x1-x0)*(y2) + x0*y1 - x1*y0 );

            //System.err.printf("alpha = %f, beta = %f, gamma = %f\n", alpha,beta,gamma);

            // Create a new fragment for the FrameBuffer's viewport.
            if (alpha >= 0 && beta >= 0 && gamma >= 0)
            {
               // Interpolate z-coordinate from the triangle's vertices to this fragment.
               double zCoord = (alpha * z0) + (beta * z1) + (gamma * z2);

               // Get the z-buffer value currently at this pixel of the viewport.
               double zBuff = fb.getDepthVP(x - 1, h - y);

               // log interesting information to standard output
               if (debug) System.out.printf("w=%d, h=%d, x=%d, y=%d, z=%f, zBuf=%f\n", w, h, x, y, zCoord, zBuff);

               // If the current fragment is in front of the pixel in the framebuffer,
               // then overwrite the pixel in the framebuffer with this fragment.
               if (zCoord > zBuff)
               {
                  // Interpolate color data from the triangle's vertices to this fragment.
                  double r = (alpha * r0) + (beta * r1) + (gamma * r2);
                  double g = (alpha * g0) + (beta * g1) + (gamma * g2);
                  double b = (alpha * b0) + (beta * b1) + (gamma * b2);
                  //System.err.printf("r = %f, g = %f, b = %f\n", r, g, b);

                   // convert from double to int
                  int ri = (int)(255 * r);
                  int gi = (int)(255 * g);
                  int bi = (int)(255 * b);
                  java.awt.Color c = new Color(ri, gi, bi);
                  fb.setPixelVP(x - 1, h - y, c, zCoord);
               }
            }
         }
      }
   }
}
